package tech.nevets.osql4j;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Field;
import java.lang.reflect.Array;
import java.sql.*;
import java.util.*;

public class SQLConnection<T> {
    private final Class<T> clazz;
    private Connection connection;
    private Statement statement;

    private final List<Field> fields;
    private final String sqlFormattedFields;
    private final String sqlFormattedFieldNames;
    private final String sqlFormattedFieldTypes;

    private static final Map<String, String> SQL_CMD_MAP = new HashMap<>();

    static {
        SQL_CMD_MAP.put("java.math.BigDecimal", "DECIMAL");
        SQL_CMD_MAP.put("boolean", "BOOLEAN");
        SQL_CMD_MAP.put("byte", "BYTE");
        SQL_CMD_MAP.put("byte[]", "BYTE");
        SQL_CMD_MAP.put("double", "FLOAT");
        SQL_CMD_MAP.put("float", "SMALLFLOAT");
        SQL_CMD_MAP.put("int", "INT");
        SQL_CMD_MAP.put("long", "INT8");
        SQL_CMD_MAP.put("short", "SMALLINT");
        SQL_CMD_MAP.put("java.lang.String", "VARCHAR");
        SQL_CMD_MAP.put("java.sql.Date", "DATE");
        SQL_CMD_MAP.put("java.sql.DateTime", "DATETIME");
        SQL_CMD_MAP.put("java.sql.Time", "TIME");
        SQL_CMD_MAP.put("java.sql.Timestamp", "DATETIME");
    }

    /**
     * This constructor is only to be used if using sqlite as a database
     * @param clazz Class of the object you want to use to structure the database
     */
    public SQLConnection(@NotNull Class<T> clazz, String dbLocation) {
        this.clazz = clazz;
        try {
            connection = DriverManager.getConnection("jdbc:sqlite:" + dbLocation);
            statement = connection.createStatement();
        } catch (SQLException e) {
            e.printStackTrace();
        }
        fields = getFields();
        sqlFormattedFields = getFieldArray();
        sqlFormattedFieldNames = getFieldNameArray();
        sqlFormattedFieldTypes = getFieldTypeArray();
        createTable();
    }

    /**
     * This constructor can be used with any db type
     * @param clazz Class of the object you want to use to structure the database
     * @param dbType MySQL (Pass credentials), or SQLite (Set credential params to null)
     * @param dbLocation SQLite (Pass location of database) or MySQL (Pass null)
     * @param dbUrl IP or FQDN to Server
     * @param dbSchema Schema to use. EX: mysql://example.com/(schema). Pass (schema)
     * @param dbUser Username credential for the Server
     * @param dbPass Password credential for the Server
     */
    public SQLConnection(@NotNull Class<T> clazz, @NotNull String dbType, @Nullable String dbLocation, @Nullable String dbUrl, @Nullable String dbSchema, @Nullable String dbUser, @Nullable String dbPass) {
        this.clazz = clazz;
        switch (dbType.toLowerCase()) {
            case "mysql" -> {
                try {
                    Class.forName("com.mysql.cj.jdbc.Driver");
                    connection = DriverManager.getConnection("jdbc:mysql://" + dbUrl + "/" + dbSchema, dbUser, dbPass);
                    statement = connection.createStatement();
                } catch (final ClassNotFoundException | SQLException e) {
                    System.out.println("Error connecting to MySQL Database...");
                    e.printStackTrace();
                }
            }
            case "sqlite" -> {
                try {
                    connection = DriverManager.getConnection("jdbc:sqlite:" + dbLocation);
                    statement = connection.createStatement();
                } catch (SQLException e) {
                    System.out.println("Error connecting to SQLite Database...");
                    e.printStackTrace();
                }
            }
        }
        fields = getFields();
        sqlFormattedFields = getFieldArray();
        sqlFormattedFieldNames = getFieldNameArray();
        sqlFormattedFieldTypes = getFieldTypeArray();
        createTable();
    }

    public void writeObject(T object) {
        try {
            List<String> fieldValues = getFieldValues(object);
            StringBuilder sql = new StringBuilder();

            sql.append("INSERT INTO ");
            sql.append(object.getClass().getSimpleName());
            sql.append("(");
            sql.append(sqlFormattedFieldNames);
            sql.append(") VALUES (");

            for (String value : fieldValues) {
                sql.append("'");
                sql.append(value);
                sql.append("', ");
            }
            sql.delete(sql.length() - 2, sql.length());
            sql.append(")");

            statement.executeUpdate(sql.toString());
        } catch (final Exception e) {
            System.out.println("Error writing object " + object.toString() + " to database...");
            e.printStackTrace();
        }
    }

    /**
     * Returns all found entries. Equivalent of doing "SELECT * FROM Table WHERE {@code column}={@code id}"
     * @param column the column in which the id will search
     * @param id the value that will be searched for
     * @return Array of all found objects
     */
    public T[] getObjectsById(String column, String id) {
        List<T> objects = new ArrayList<>();
        try {
            String sql = "SELECT * FROM " + clazz.getSimpleName() + " WHERE " + column + "=" + id + ";";
            ResultSet results = statement.executeQuery(sql);
            while (results.next()) {
                objects.add(getFromDatabase(results));
            }
        } catch (final Exception e) {
            System.out.println("Error parsing object from database...");
            e.printStackTrace();
        }
        @SuppressWarnings("unchecked")
        T[] objectArray = (T[]) Array.newInstance(this.clazz, objects.size());
        return objectArray;
    }

    public T[] getAllObjects() {
        List<T> objects = new ArrayList<>();
        try {
            String sql = "SELECT * FROM " + clazz.getSimpleName() + ";";
            ResultSet results = statement.executeQuery(sql);
            while (results.next()) {
                objects.add(getFromDatabase(results));
            }
        } catch (final SQLException e) {
            System.out.println("Error parsing objects from database...");
            e.printStackTrace();
        }
        @SuppressWarnings("unchecked")
        T[] objectArray = (T[]) Array.newInstance(this.clazz, objects.size());
        return objectArray;
    }

    public DatabaseMetaData getMetadata() {
        try {
            return connection.getMetaData();
        } catch (SQLException e) {
            System.out.println("Error fetching metadata...");
            e.printStackTrace();
            return null;
        }
    }

    public void close() {
        try {
            if (connection != null && !connection.isClosed()) {
                connection.close();
            }
        } catch (SQLException e) {
            System.out.println("Error closing SQL Connection");
            e.printStackTrace();
        }
    }

    private void createTable() {
        try {
            String sql = "CREATE TABLE IF NOT EXISTS " +
                    clazz.getSimpleName() +
                    " (id INT PRIMARY KEY, " +
                    sqlFormattedFields +
                    " );";
            statement.executeUpdate(sql);
        } catch (SQLException e) {
            System.out.println("Error creating new table " + clazz.getSimpleName() + "...");
            e.printStackTrace();
        }
    }

    private T getFromDatabase(ResultSet results) {
        try {
            T object = clazz.getConstructor().newInstance();
            for (int i = 0; i < fields.size(); i++) {
                String fieldName = fields.get(i).getName();
                String fieldType = fields.get(i).getType().getName();
                Field objectField = object.getClass().getDeclaredField(fieldName);
                objectField.setAccessible(true);

                switch (fieldType) {
                    case "boolean", "java.lang.Boolean" -> objectField.setBoolean(object, results.getBoolean(i + 2));
                    case "byte", "java.lang.Byte" -> objectField.setByte(object, results.getByte(i + 2));
                    case "byte[]", "java.lang.Byte[]" -> objectField.set(object, results.getBytes(i + 2));
                    case "double", "java.lang.Double" -> objectField.setDouble(object, results.getFloat(i + 2));
                    case "float", "java.lang.Float" -> objectField.setFloat(object, results.getFloat(i + 2));
                    case "int", "java.lang.Integer" -> objectField.setInt(object, results.getInt(i + 2));
                    case "long", "java.lang.Long" -> objectField.setLong(object, results.getLong(i + 2));
                    case "short", "java.lang.Short" -> objectField.setShort(object, results.getShort(i + 2));
                    case "java.lang.String" -> objectField.set(object, results.getString(i + 2));
                    case "java.math.BigDecimal" -> objectField.set(object, results.getBigDecimal(i + 2));
                    case "java.sql.Date" -> objectField.set(object, results.getDate(i + 2));
                    case "java.sql.Time" -> objectField.set(object, results.getTime(i + 2));
                    case "java.sql.TimeStamp" -> objectField.set(object, results.getTimestamp(i + 2));
                }
            }
            return object;
        } catch (final Exception e) {
            System.out.println("Error parsing object from database result");
            e.printStackTrace();
            return null;
        }
    }

    private List<Field> getFields() {
        List<Field> fieldsList = new ArrayList<>(List.of(clazz.getDeclaredFields()));
        for (Class<?> c = clazz; c != null; c = c.getSuperclass()) {
            fieldsList.addAll(Arrays.asList(c.getDeclaredFields()));
        }
        return fieldsList;
    }

    private String getFieldArray() {
        StringBuilder sql = new StringBuilder();
        for (Field field : fields) {
            sql.append(field.getName());
            sql.append(" ");
            sql.append(SQL_CMD_MAP.get(field.getType().getName()));
            sql.append(", ");
        }
        sql.delete(sql.length() - 2, sql.length());
        return sql.toString();
    }

    private String getFieldNameArray() {
        StringBuilder sql = new StringBuilder();
        for (Field field : fields) {
            sql.append(field.getName());
            sql.append(", ");
        }
        sql.delete(sql.length() - 2, sql.length());
        return sql.toString();
    }

    private String getFieldTypeArray() {
        StringBuilder sql = new StringBuilder();
        for (Field field : fields) {
            sql.append(SQL_CMD_MAP.get(field.getType().getName()));
            sql.append(", ");
        }
        sql.delete(sql.length() - 2, sql.length());
        return sql.toString();
    }

    private List<String> getFieldValues(T object) {
        List<String> values = new ArrayList<>();
        for (Field field : fields) {
            try {
                field.setAccessible(true);
                values.add(field.get(object).toString());
            } catch (IllegalAccessException e) {
                System.out.println("Error getting field values...");
                e.printStackTrace();
            }
        }
        return values;
    }
}
